/*
    This file is part of an implementation of the UCSD I.5 psuedo-machine.
    Copyright (C) 1994,1995,2006 Richard. W. Stearn

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   Email: richard@rns-stearn.demon.co.uk

    Inspiration and some of the code comes from the UCSD II.0 emulator for
    the Apple ][ P-System by Mario Klebsch <http://www.klebsch.de>
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include "pme.h"
#include "ucsd_consts.h"
#include "ucsd_syscom.h"
#include "interpreter.h"
#include "memory.h"
#include "unit_io.h"

#define WORD_SZ		16
#define SET_SZ		512
#define MAX_STR_LEN	255

#define	SLDL0		216
#define	SLDO0		232
#define SIND0		248

#define CMP_REAL	2
#define CMP_STR		4
#define CMP_BOOL	6
#define CMP_SET		8
#define CMP_BYTES	10
#define CMP_WORDS	12

#define FALSE		0
#define TRUE		1

#define GREATER_THAN	1
#define EQUAL_TO	0
#define LESS_THAN	-1
#define NOT_EQUAL	1

#define POINTER_CHECK( pointer ) if ( (pointer) == NIL ) { xeqerr( ENILPTR, "ptr chk" ); return NIL; }

#define MSCW_RD( pointer, offset ) ( mem_rd_w( WORD_ADDR( (pointer), (offset) ) )) 
#define MSCW_WR( pointer, offset, value ) ( mem_wr_w( WORD_ADDR( pointer, offset ), value ) )

#define WORD_ALIGN( ipc ) ( ( (ipc) + 1 ) & ( ~1 ) )

#define MIN( a, b ) ( ( (a) < (b) ) ? (a) : (b) )
#define MAX( a, b ) ( ( (a) > (b) ) ? (a) : (b) )

extern int dflag;

/* p-m/c registers */
extern word sp;
word np;
word mp;
word seg;
word jtab;
word base;
word ipc;
word basemp;

/* SYSCOM	; Interpreter and pascalsystem communication area */
struct syscom_rec *syscom;

struct int_seg_entry int_seg_tbl[ MAXSEG + 1 ];

void xeqerr( word err, char *detail );

/* -------------------------------------------------------------------------- */
/* symbols and table for id_search */
#define SYM_IDENT	0
#define SYM_DO		6
#define SYM_TO		7
#define SYM_DOWNTO	8
#define SYM_END		9
#define SYM_UNTIL	10
#define SYM_OF		11
#define SYM_THEN	12
#define SYM_ELSE	13
#define SYM_BEGIN	19
#define SYM_IF		20
#define SYM_CASE	21
#define SYM_REPEAT	22
#define SYM_WHILE	23
#define SYM_FOR		24
#define SYM_WITH	25
#define SYM_GOTO	26
#define SYM_LABEL	27
#define SYM_CONST	28
#define SYM_TYPE	29
#define SYM_VAR		30
#define SYM_PROC	31
#define SYM_FUNC	32
#define SYM_PROG	33
#define SYM_FORWARD	34
#define SYM_NOT		38
#define SYM_MULOP	39
#define SYM_ADDOP	40
#define SYM_RELOP	41
#define SYM_SET		42
#define SYM_PACKED	43
#define SYM_ARRAY	44
#define SYM_RECORD	45
#define SYM_FILE	46
#define SYM_USES	49
#define SYM_UNIT	50
#define SYM_INTER	51
#define SYM_IMPLE	52
#define SYM_EXTERNL	53

#define OP_AND		2
#define OP_IDIV		3
#define OP_IMOD		4
#define OP_OR		7
#define OP_IN		14
#define OP_NOOP		15

#define TOKEN_LEN	8

struct id_entry
	{
	byte *token;
	byte sym;
	byte op;
	};

static struct id_entry id_table[] =
	{
	{ "AND     ", SYM_MULOP,	OP_AND   },
	{ "ARRAY   ", SYM_ARRAY,	OP_NOOP  },
	{ "BEGIN   ", SYM_BEGIN,	OP_NOOP  },
	{ "CASE    ", SYM_CASE,		OP_NOOP  },
	{ "CONST   ", SYM_CONST,	OP_NOOP  },
	{ "DIV     ", SYM_MULOP,	OP_IDIV  },
	{ "DO      ", SYM_DO,		OP_NOOP  },
	{ "DOWNTO  ", SYM_DOWNTO,	OP_NOOP  },
	{ "ELSE    ", SYM_ELSE,		OP_NOOP  },
	{ "END     ", SYM_END,		OP_NOOP  },
	{ "EXTERNAL", SYM_EXTERNL,	OP_NOOP  },
	{ "FILE    ", SYM_FILE,		OP_NOOP  },
	{ "FOR     ", SYM_FOR,		OP_NOOP  },
	{ "FORWARD ", SYM_FORWARD,	OP_NOOP  },
	{ "FUNCTION", SYM_FUNC,		OP_NOOP  },
	{ "GOTO    ", SYM_GOTO,		OP_NOOP  },
	{ "IF      ", SYM_IF,		OP_NOOP  },
	{ "IMPLEMEN", SYM_IMPLE,	OP_NOOP  },
	{ "IN      ", SYM_RELOP,	OP_IN    },
	{ "INTERFAC", SYM_INTER,	OP_NOOP  },
	{ "LABEL   ", SYM_LABEL,	OP_NOOP  },
	{ "MOD     ", SYM_MULOP,	OP_IMOD  },
	{ "NOT     ", SYM_NOT,		OP_NOOP  },
	{ "OF      ", SYM_OF,		OP_NOOP  },
	{ "OR      ", SYM_ADDOP,	OP_OR    },
	{ "PACKED  ", SYM_PACKED,	OP_NOOP  },
	{ "PROCEDUR", SYM_PROC,		OP_NOOP  },
	{ "PROGRAM ", SYM_PROG,		OP_NOOP  },
	{ "RECORD  ", SYM_RECORD,	OP_NOOP  },
	{ "REPEAT  ", SYM_REPEAT,	OP_NOOP  },
	{ "SEGMENT ", SYM_PROG,		OP_NOOP  },
	{ "SET     ", SYM_SET,		OP_NOOP  },
	{ "THEN    ", SYM_THEN,		OP_NOOP  },
	{ "TO      ", SYM_TO,		OP_NOOP  },
	{ "TYPE    ", SYM_TYPE,		OP_NOOP  },
	{ "UNIT    ", SYM_UNIT,		OP_NOOP  },
	{ "UNTIL   ", SYM_UNTIL,	OP_NOOP  },
	{ "USES    ", SYM_USES,		OP_NOOP  },
	{ "VAR     ", SYM_VAR,		OP_NOOP  },
	{ "WHILE   ", SYM_WHILE,	OP_NOOP  },
	{ "WITH    ", SYM_WITH,		OP_NOOP  },
	{ NULL      , 0,		0        }
	};

/* -------------------------------------------------------------------------- */
void mv_left( word dst_addr, word src_addr, word len )
{
if ( ! ( len & 0x8000 ) )
	while ( len )
		{
		mem_wr_b( dst_addr, mem_rd_b( src_addr ) );
		len--;
		dst_addr++;
		src_addr++;
		}
}

/* -------------------------------------------------------------------------- */
void mv_right( word dst_addr, word src_addr, word len )
{
src_addr += len;
dst_addr += len;
if ( ! ( len & 0x8000 ) )
	while ( len )
		{
		len--;
		dst_addr--;
		src_addr--;
		mem_wr_b( dst_addr, mem_rd_b( src_addr ) );
		}
}

/* -------------------------------------------------------------------------- */
word fetch_b()
{
byte b;

b = mem_rd_b( ipc );
ipc += 1;
return b;
}

/* -------------------------------------------------------------------------- */
word get_big()
{
word big;

big = fetch_b();
if ( big & 0x80 )
	big = ( ( big & 0x7f ) << 8 ) | fetch_b();

return big;
}

/* -------------------------------------------------------------------------- */
word fetch_w()
{
word wrd;

wrd = mem_rd_w( ipc );
ipc += 2;
return wrd;
}

/* -------------------------------------------------------------------------- */
/* Traverse up the static links "count" times. */
word traverse_links( byte count )
{
word s_link;

for ( s_link = mp; count; count-- )
	s_link = MSCW_RD( s_link, MS_STAT  );
return s_link;
}

/* -------------------------------------------------------------------------- */
/* Returns a pointer to a global variable. */
word global_addr( word offset )
{
return( WORD_ADDR( base, MS_VAR + offset ) );
}

/* -------------------------------------------------------------------------- */
/* Returns a pointer to a variable in an intermediate procedure. */
word intermediate_addr( word offset, byte count )
{
return( WORD_ADDR( traverse_links( count ), MS_VAR + offset ) );
}

/* -------------------------------------------------------------------------- */
/* Returns a pointer to a local variable. */
word local_addr( word offset )
{
return( WORD_ADDR( mp, MS_VAR + offset ) );
}

/* -------------------------------------------------------------------------- */
word boolean( word word_val )
{
return( word_val ? 1 : 0 );
}

/* -------------------------------------------------------------------------- */
/* Resolve a self-relative pointer. */
word self_rel_ptr( word addr )
{
return( addr - mem_rd_w( addr ) );
}

/* -------------------------------------------------------------------------- */
/* Pointer to the first instruction of a procedure */
word entry_ipc( word jtabp )
{
POINTER_CHECK( jtabp );
return ( self_rel_ptr( WORD_ADDR( jtabp, JTAB_ENTRY_IPC ) ) );
}

/* -------------------------------------------------------------------------- */
/* Pointer to the exit code of a procedure. */
word exit_ipc( word jtabp )
{
POINTER_CHECK( jtabp );
return ( self_rel_ptr( WORD_ADDR( jtabp, JTAB_EXIT_IPC ) ) );
}

/* -------------------------------------------------------------------------- */
/* calculate the target address of a jump operation.
   Positive displacements perform relative jumps,
   negative displacements are indices into the jump table. */
word jump( signed_byte disp )
{
if ( disp >= 0 )
	return( ipc + disp );

disp = -disp;
return( entry_ipc( jtab )
	- ( mem_rd_w( jtab - disp ) + disp )
	+ mem_rd_w( WORD_ADDR( jtab, JTAB_ENTRY_IPC ) )
	+ 2
	);
}

/* -------------------------------------------------------------------------- */
/* string compare: returns -1 if s1 < s2, 0 if s1 == s2, 1 if s1 > s2 */
signed_word cmp_str( word s1, word s2 )
{
byte len1;
byte len2;
byte min_len;
byte index;
byte ch1;
byte ch2;

len1 = mem_rd_b( s1 );
len2 = mem_rd_b( s2 );

min_len  = MIN( len1, len2 );
for ( index = 0; index < min_len; index++ )
	{
	ch1 = mem_rd_b( s1 + index + 1 );
	ch2 = mem_rd_b( s2 + index + 1 );

	if ( ch1 < ch2 )
		return( LESS_THAN );

	if ( ch1 > ch2 )
		return( GREATER_THAN );
	}

if ( len1 < len2 )
	return( LESS_THAN );

if ( len1 > len2 )
	return( GREATER_THAN );

return( EQUAL_TO );
}

/* -------------------------------------------------------------------------- */
/* pads end of set with zeros */
void adj_s( set_t *set, word size )
{
int count;

for ( count = set->size; count < size; count++ )
	set->data[ count ] = 0;
set->size = size;
}

/* -------------------------------------------------------------------------- */
/* returns 1 if set2 is equal to set1 */
int cmp_set(set_t *set1, set_t *set2)
{
int size;
int count;

size = MAX( set1->size, set2->size );
if ( set1->size < size )
	adj_s( set1, size );
if ( set2->size < size )
	adj_s( set2, size );

for ( count = 0; count < size; count++ )
	if ( set1->data[ count ] != set2->data[ count ] )
		return( TRUE );
return( FALSE );
}

/* -------------------------------------------------------------------------- */
/* returns 1 if set2 is subset of set1 */
int is_subset( set_t *set1, set_t *set2 )
{
int size;
int index;

size = set2->size;
if ( set1->size < size )
	adj_s( set1, size );

for ( index = 0; index < size; index++ )
	if  ( ( set1->data[ index ] & set2->data[ index ] ) != set2->data[ index ] )
		return( FALSE );
return( TRUE );
}

/* -------------------------------------------------------------------------- */
/* byte array compare: returns -1 if ba1 < ba2, 0 if ba1 == ba2, 1 if ba1 > ba2 */
signed_word cmp_bytes( word ba1, word ba2, word len )
{
word index;
byte ch1;
byte ch2;

for ( index = 0; index < len; index++ )
	{
	ch1 = mem_rd_b( ba1 + index );
	ch2 = mem_rd_b( ba2 + index );

	if ( ch1 < ch2 )
		return( LESS_THAN );

	if ( ch1 > ch2 )
		return( GREATER_THAN );
	}

return( EQUAL_TO );
}

/* -------------------------------------------------------------------------- */
/* word array compare: returns 0 if wa1 == wa2, 1 if wa1 != wa2 */
signed_word cmp_words( word wa1, word wa2, word len )
{
word	index;

for ( index = 0; index < len; index++ )
	if ( mem_rd_w( WORD_ADDR( wa1, index ) ) != mem_rd_w( WORD_ADDR( wa2, index ) ) )
		return( NOT_EQUAL );
return( EQUAL_TO );
}

/* -------------------------------------------------------------------------- */
/* Return the segment number for the given segment pointer */
byte seg_number( word segp )
{
return( mem_rd_w( segp ) & 0xff );
}

/* -------------------------------------------------------------------------- */
/* Returns the number of procedures in the segment */
byte seg_num_proc( word segp )
{
POINTER_CHECK( segp );
return( mem_rd_w( segp ) >> 8 );
}

/* -------------------------------------------------------------------------- */
/* load a segment. If a data segment is to be loaded, just allocate storage on the stack */
int load_segment( byte seg_num )
{
word seg_unit;
word seg_block;
word seg_size;
int loaded_code;

loaded_code = 0;
if ( int_seg_tbl[ seg_num ].ref_count <= 0 )
	{
	seg_unit  = syscom->segtbl[ seg_num ].code_unit;
	seg_block = syscom->segtbl[ seg_num ].disk_addr;
	seg_size  = syscom->segtbl[ seg_num ].code_len;
	if ( seg_size <= 0 )
		{
		xeqerr( ENOPROC, "load_segment" );
		return 0;
		}
	if ( seg_block <= 0 )
		{
		xeqerr( ENOPROC, "load_segment" );
		return 0;
		}
	sp = sp - seg_size;
	syscom->iorslt = unit_read( seg_unit, sp, seg_size, seg_block, 0 );
	if ( syscom->iorslt != 0 )
		{
		xeqerr( ESYIOER, "load_segment" );
		return 0;
		}
	int_seg_tbl[ seg_num ].code_addr = sp;
	int_seg_tbl[ seg_num ].ref_count = 0;
	loaded_code = 1;
	}
int_seg_tbl[ seg_num ].ref_count++;
return loaded_code;
}

/* -------------------------------------------------------------------------- */
void unload_segment( byte seg_num )
{
if ( seg_num == 0 )
	{
	return;
	}

int_seg_tbl[ seg_num ].ref_count--;
if ( int_seg_tbl[ seg_num ].ref_count <= 0 )
	{
	int_seg_tbl[ seg_num ].code_addr = NIL;
	int_seg_tbl[ seg_num ].ref_count = 0;
	}
}

/* -------------------------------------------------------------------------- */
/* check for a gap between heap and stack */
void stack_check()
{
if ( np >= sp )
	{
	syscom->gdirp = NIL;
	sp = 0x8000;
	np = 0x6200;
	xeqerr( ESTKOVR, "stack_check" );
	}
}

/* -------------------------------------------------------------------------- */
/* Returns a pointer to the jump table of a the procedure in the segment */
word get_jtab( word segp, byte proc_num )
{
POINTER_CHECK( segp );
if ( proc_num < 1 )
	{
	xeqerr( ENOPROC, "get_jtab: proc_num < 1" );
	return 0;
	}
if ( proc_num > seg_num_proc( segp ) )
	{
	xeqerr( ENOPROC, "get_jtab: proc_num > segs in proc" );
	return 0;
	}
return( self_rel_ptr( WORD_ADDR( segp, -proc_num ) ) );
}

/* -------------------------------------------------------------------------- */
/* Returns the procedure number of the current procedure */
signed_byte proc_number( word jtabp )
{
POINTER_CHECK( jtabp );
return ( mem_rd_w( jtabp ) & 0xff );
}

/* -------------------------------------------------------------------------- */
/* Returns the lex level of the current procedure */
signed_byte lex_level( word jtabp )
{
POINTER_CHECK( jtabp );
return ( mem_rd_w( jtabp ) >> 8 );
}

/* -------------------------------------------------------------------------- */
void native_proc( word jtabp )
{
xeqerr( ENOTIMP, "native proc" );
}

/* -------------------------------------------------------------------------- */
/* Calculates the static link pointer for a procedure */
word static_link( word segp, byte proc_num )
{
word new_jtab;

new_jtab = get_jtab( segp, proc_num );

if ( ! proc_number( new_jtab ) )
	return( NIL );

return( traverse_links( lex_level( jtab ) - lex_level( new_jtab ) + 1 ) );
}

/* -------------------------------------------------------------------------- */
/* Clear the global directory pointer. */
void clear_gdirp()
{
if ( syscom->gdirp != NIL )
	{
	np = syscom->gdirp;
	syscom->gdirp = NIL;
	}
}

/* -------------------------------------------------------------------------- */
/* call a procedure. It builds the stack frame for the new procedure
   and sets up all the p-machine. registers */
int call( word new_seg, byte proc_num, word static_link, byte code_loaded )
{
word new_mp;
word new_jtab;
word data_size;
word param_size;
word param_ptr;

if ( seg_number( new_seg ) > MAXSEG )
	{
	xeqerr( ENOPROC, "call" );
	return 1;
	}

new_jtab = get_jtab( new_seg, proc_num );
POINTER_CHECK( new_jtab );

if ( ! proc_number( new_jtab ) )
	{
	native_proc( new_jtab );
	return( 1 );
	}

data_size  = mem_rd_w( WORD_ADDR( new_jtab, JTAB_DATA_SZ ) );
param_size = mem_rd_w( WORD_ADDR( new_jtab, JTAB_PARAM_SZ ) );

param_ptr  = sp;
sp = BYTE_ADDR( sp, -( data_size ) );
if ( code_loaded )
	{
	param_ptr = new_seg + 2;
	sp = BYTE_ADDR( sp, -( param_size ) );
	}
mv_left( sp, param_ptr, param_size );

/* build MSCW */
push_w( param_ptr );	/* MS_SP   */
push_w( ipc );		/* MS_IPC  */
push_w( seg );		/* MS_SEG  */
push_w( jtab );		/* MS_JTAB */
push_w( mp );		/* MS_DYN  */
push_w( static_link );	/* MS_STAT */
new_mp = sp;

if ( ( lex_level( new_jtab ) <= 0 ) /* a base procedure */
	|| ( ( seg_number( seg ) != 0 ) && ( new_seg == 0 ) ) /* a procedure in seg 0 */
	)
	{
	push_w( base );
	base = new_mp;
	syscom->stkbase = base;
	}

syscom->lastmp = new_mp;
syscom->seg    = new_seg;
syscom->jtab   = new_jtab;

mp   = new_mp;
seg  = new_seg;
jtab = new_jtab;
ipc  = entry_ipc( new_jtab );

stack_check();

return( 0 );
}

/* -------------------------------------------------------------------------- */
void ret( byte ret_size )
{
word old_mp;
byte old_seg;
byte old_proc;
word old_sp;
word dummy;
word ret_val_ptr;
word data_size;
word param_size;

sp       = mp;
old_mp   = mp;
old_seg  = seg_number( seg );
old_proc = proc_number( jtab );

data_size  = mem_rd_w( WORD_ADDR( jtab, JTAB_DATA_SZ ) );
param_size = mem_rd_w( WORD_ADDR( jtab, JTAB_PARAM_SZ ) );
ret_size   = ret_size + ret_size;

ret_val_ptr = local_addr( 1 );

dummy = pop_w(); /* MS_STAT */
mp    = pop_w(); /* MS_DYN  */
jtab  = pop_w(); /* MS_JTAB */
seg   = pop_w(); /* MS_SEG  */
ipc   = pop_w(); /* MS_IPC  */
old_sp= pop_w(); /* MS_SP   */

syscom->lastmp = mp;
syscom->seg    = seg;
syscom->jtab   = jtab;

mv_right( old_sp + param_size - ret_size, ret_val_ptr, ret_size );
sp = old_sp + param_size - ret_size;

if ( old_seg != 0 )
	if ( old_seg != seg_number( seg ) )
		{
		unload_segment( old_seg );
		}

stack_check();
}

/* -------------------------------------------------------------------------- */
void id_search( word reserved_word_ptr, word sym_buffer_ptr )
{
word offset;
byte token[ TOKEN_LEN + 1 ];
byte ch;
int index;

offset = mem_rd_w( sym_buffer_ptr );
memset( token, ' ', sizeof( token ) );

index = 0;
while ( 1 )
	{
	ch = toupper( mem_rd_b( reserved_word_ptr + offset ) );
	if ( ch != '_' )
		{
		if ( ! isalpha( ch ) && ! isdigit( ch ) )
			break;
		if ( index < TOKEN_LEN )
			{
			token[ index ] = ch;
			index++;
			}
		}
	offset++;
	}
offset--;
token[ TOKEN_LEN ] = '\0';

mem_wr_w( sym_buffer_ptr, offset );

index = 0;
while ( id_table[ index ].token != NULL )
	{
	if ( strncmp( id_table[ index ].token, token, TOKEN_LEN ) == 0 )
		{
		mem_wr_w( WORD_ADDR( sym_buffer_ptr, 1 ), id_table[ index ].sym );
		mem_wr_w( WORD_ADDR( sym_buffer_ptr, 2 ), id_table[ index ].op );
		return;
		}
	index++;
	}

mem_wr_w( WORD_ADDR( sym_buffer_ptr, 1 ), SYM_IDENT );
mem_wr_w( WORD_ADDR( sym_buffer_ptr, 2 ), OP_NOOP );
for ( index = 0; index < TOKEN_LEN; index++ )
	mem_wr_b( WORD_ADDR( sym_buffer_ptr, 3 ) + index, token[ index ] );
}

/* -------------------------------------------------------------------------- */
word tree_search( word token, word result_ptr, word node_ptr )
{
word link;
signed_word cmp;

while ( 1 )
	{
	cmp = (signed_word) cmp_bytes( token, node_ptr, 8 );
	if ( cmp < 0 )
		{
		link = mem_rd_w( WORD_ADDR( node_ptr, 5 ) );
		if ( link != NIL )
			node_ptr = link;			/* follow right link */
		else
			{
			mem_wr_w( result_ptr, node_ptr );
			return( 0xffff );
			}
		}
	else
		if ( cmp > 0 )
			{
			link = mem_rd_w( WORD_ADDR( node_ptr, 4 ) );
			if ( link != NIL )
				node_ptr = link;		/* follow left link */
			else
				{
				mem_wr_w( result_ptr, node_ptr );
				return( 1 );
				}
			}
		else
			{
			mem_wr_w( result_ptr, node_ptr );
			return( 0 );
			}
	}
}

/* ========================================================================== */
/* logic p-codes */

/* -------------------------------------------------------------------------- */
/* AND */
void and()
{
push_w( pop_w() & pop_w() );
}

/* -------------------------------------------------------------------------- */
/* IOR */
void ior()
{
push_w( pop_w() | pop_w() );
}

/* -------------------------------------------------------------------------- */
/* NOT */
void not()
{
push_w( ~( pop_w() ) );
}

/* ========================================================================== */
/* load constant */

/* -------------------------------------------------------------------------- */
/* SDLC0..127 */
/* -------------------------------------------------------------------------- */
/* LDCI       load constant word */
void ldci()
{
push_w( fetch_w() );
}

/* -------------------------------------------------------------------------- */
/* LDCN       load constant nil */
void ldcn()
{
push_w( NIL );
}

/* ========================================================================== */
/* integer arith p-codes */

/* -------------------------------------------------------------------------- */
/* ABI        absolute value of integer */
void abi()
{
push_w( abs( (signed_word) pop_w() ) );
}

/* -------------------------------------------------------------------------- */
/* ADI        add integers */
void adi()
{
push_w( (signed_word) pop_w() + (signed_word) pop_w() );
}

/* -------------------------------------------------------------------------- */
/* SBI        subtract integers */
void sbi()
{
word tos;

tos = pop_w();
push_w( (signed_word) pop_w() - (signed_word) tos );
}

/* -------------------------------------------------------------------------- */
/* DVI        divide integers */
void dvi()
{
word tos;

tos = pop_w();
if ( ! tos )
	{
	xeqerr( EDIVZER, "dvi" );
	return;
	}
push_w( (signed_word) pop_w() / (signed_word) tos );
}

/* -------------------------------------------------------------------------- */
/* MOD        modulo integers */
void modi()
{
word tos;

tos = pop_w(); 
if ( ! tos )
	{
	xeqerr( EDIVZER, "modi" );
	return;
	}
push_w( (signed_word) pop_w() % (signed_word) tos );
}

/* -------------------------------------------------------------------------- */
/* MPI        multiply integers */
void mpi()
{
push_w( (signed_word) pop_w() * (signed_word) pop_w() );
}

/* -------------------------------------------------------------------------- */
/* SQI        square integers */
void sqi()
{
word tos;

tos = pop_w();
push_w( (signed_word) tos * (signed_word) tos );
}

/* -------------------------------------------------------------------------- */
/* NGI        negate integers */
void ngi()
{
push_w( -( (signed_word) pop_w() ) );
}

/* -------------------------------------------------------------------------- */
/* CHK        check against sub-range bounds */
void chk()
{
word upper;
word lower;
word index;

upper = pop_w();
lower = pop_w();
index = pop_w();
push_w( index );
if ( ( index > upper ) || ( index < lower ) )
	xeqerr( EINVNDX, "chk" );
}

/* ========================================================================== */
/* real arith p-codes */

/* -------------------------------------------------------------------------- */
/* ABR        take absolute value of real */
void abr()
{
push_r( fabs( pop_r() ) );
}

/* -------------------------------------------------------------------------- */
/* ADR        add reals */
void adr()
{
push_r( pop_r() + pop_r() );
}

/* -------------------------------------------------------------------------- */
/* MPR        multiply reals */
void mpr()
{
push_r( pop_r() * pop_r() );
}

/* -------------------------------------------------------------------------- */
/* NGR        negate reals */
void ngr()
{
push_r( -( pop_r() ) );
}

/* -------------------------------------------------------------------------- */
/* SBR        subtract reals */
void sbr()
{
real tos;

tos = pop_r();
push_r( pop_r() - tos );
}

/* -------------------------------------------------------------------------- */
/* DVR        divide reals */
void dvr()
{
real tos;

tos = pop_r();
if ( tos == 0 )
	{
	xeqerr( EDIVZER, "dvr" );
	return;
	}
push_r( pop_r() / tos );
}

/* -------------------------------------------------------------------------- */
/* SQR        square reals */
void sqr()
{
real tos;

tos = pop_r();
push_r( tos * tos );
}

/* -------------------------------------------------------------------------- */
/* FLO        float next to top of stack */
void flo()
{
real tos;

tos = pop_r();
push_r( pop_w() );
push_r( tos );
}

/* -------------------------------------------------------------------------- */
/* FLT        float top of stack */
void flt()
{
push_r( (signed_word) pop_w() );
}

/* ========================================================================== */
/* multiple word vars (sets and reals) p-codes */

/* -------------------------------------------------------------------------- */
/* LDC        load multiple word constant */
void ldc()
{
word count;

count = fetch_b();
ipc = WORD_ALIGN( ipc );
while ( count )
	{
	push_w( fetch_w() );
	count--;
	}
}

/* -------------------------------------------------------------------------- */
/* LDM        load multiple words */
void ldm()
{
word len;
word src_addr;

len = fetch_b();
src_addr = pop_w();
while ( len )
	{
	len--;
	push_w( mem_rd_w( WORD_ADDR( src_addr, len ) ) );
	}
}

/* -------------------------------------------------------------------------- */
/* STM        store multiple words */
void stm()
{
word len;
word dst_addr;

len = fetch_b();
dst_addr = mem_rd_w( WORD_ADDR( sp, len ) );
while ( len )
	{
	len--;
	mem_wr_w( dst_addr, pop_w() );
	dst_addr = WORD_ADDR( dst_addr, 1 );
	}
pop_w();
}

/* ========================================================================== */
/* local load & store p-codes */

/* -------------------------------------------------------------------------- */
/* LLA        load local address */
void lla()
{
push_w( local_addr( get_big() ) );
}

/* -------------------------------------------------------------------------- */
/* LDL        load local word */
void ldl()
{
push_w( mem_rd_w( local_addr( get_big() ) ) );
}

/* -------------------------------------------------------------------------- */
/* STL        store local word */
void stl()
{
mem_wr_w( local_addr( get_big() ), pop_w() );
}

/* -------------------------------------------------------------------------- */
/* SLDL 0..15 short load local word */
void sldl( word opcode )
{
push_w( mem_rd_w( local_addr( opcode - SLDL0 + 1) ) );
}

/* ========================================================================== */
/* intermediate load & store p-codes */

/* -------------------------------------------------------------------------- */
/* LDA        load intermediate address */
void lda()
{
word tos;

tos = fetch_b();
push_w( intermediate_addr( get_big(), tos ));
}

/* -------------------------------------------------------------------------- */
/* LOD        load intermediate word */
void lod()
{
word offset;

offset = fetch_b();
push_w( mem_rd_w( intermediate_addr( get_big(), offset )));
}

/* -------------------------------------------------------------------------- */
/* STR        store intermediate word */
void str()
{
word offset;

offset = fetch_b();
mem_wr_w( intermediate_addr( get_big(), offset ), pop_w() );
}

/* ========================================================================== */
/* global load & store p-codes */

/* -------------------------------------------------------------------------- */
/* LAO        load global address */
void lao()
{
push_w( global_addr( get_big() ) );
}

/* -------------------------------------------------------------------------- */
/* LDO        load global word */
void ldo()
{
push_w( mem_rd_w( global_addr( get_big() ) ) );
}

/* -------------------------------------------------------------------------- */
/* SRO        store global word */
void sro()
{
mem_wr_w( global_addr( get_big() ), pop_w() );
}

/* -------------------------------------------------------------------------- */
/* SLDO 0..15 short load global word */
void sldo( word opcode )
{
push_w( mem_rd_w( global_addr( opcode - SLDO0 + 1) ) );
}

/* ========================================================================== */
/* procedure/function call p-codes */

/* -------------------------------------------------------------------------- */
/* CLP        call local procedure */
void clp()
{
call( seg, fetch_b(), mp, 0 );
}

/* -------------------------------------------------------------------------- */
/* CIP        call intermediate procedure */
void cip()
{
word proc_num;

proc_num = fetch_b();
call( seg, proc_num, static_link( seg, proc_num ), 0 );
}

/* -------------------------------------------------------------------------- */
/* CGP        call global procedure */
void cgp()
{
call( seg, fetch_b(), base, 0 );
}

/* -------------------------------------------------------------------------- */
void cbp()
/* CBP        call base procedure */
{
call( seg, fetch_b(), basemp, 0 );
}

/* -------------------------------------------------------------------------- */
/* CXP        call external procedure */
void cxp()
{
word seg_num;
word proc_num;
word new_seg;
byte code_loaded;

seg_num = fetch_b();
proc_num = fetch_b();

code_loaded = 0;
if ( seg_num )
	code_loaded = load_segment( seg_num );
new_seg = int_seg_tbl[ seg_num ].code_addr + syscom->segtbl[ seg_num ].code_len - 2;

if ( call( new_seg, proc_num, static_link( new_seg, proc_num ), code_loaded ) )
	unload_segment( seg_num );
}

/* -------------------------------------------------------------------------- */
/* CSP        call standard procedure */
void csp()
{
switch( fetch_b() )
	{
	case CSP_IOC:
		if ( syscom->iorslt )
			xeqerr( EUIOERR, "csp_ioc" );
		break;
	case CSP_NEW:
		{
		word size;

		clear_gdirp();
		size = pop_w();
		mem_wr_w( pop_w(), np );
		np = WORD_ADDR( np, size );
		stack_check();
		}
		break;
	case CSP_MVL:
		{
		word len;
		word dst;
		word src;

		len = pop_w();
		dst = pop_w();
		src = pop_w();
		mv_left( dst, src, len );
		}
		break;
	case CSP_MVR:
		{
		word len;
		word dst;
		word src;

		len = pop_w();
		dst = pop_w();
		src = pop_w();
		mv_right( dst, src, len );
		}
		break;
	case CSP_XIT:
		{
		word proc_num;
		word seg_num;
		word exit_mp;
		word exit_seg;
		word exit_jtab;

		proc_num = pop_w();
		seg_num = pop_w();
		exit_mp = mp;
		exit_seg = seg;
		exit_jtab = jtab;
		ipc = exit_ipc( exit_jtab );
		while ( ( proc_number( exit_jtab ) != proc_num ) ||
			( seg_number( exit_seg ) != seg_num  ) )
			{
			if ( exit_mp == basemp )
				{
				xeqerr( ENOEXIT, "csp_xit: exit_mp" );
				return;
				}

			exit_jtab = MSCW_RD( exit_mp, MS_JTAB );
			if ( exit_jtab == 0 )
				{
				xeqerr( ENOEXIT, "csp_xit: exit_jtab" );
				return;
				}

			exit_seg  = MSCW_RD( exit_mp, MS_SEG );
			if ( exit_seg == 0 )
				{
				xeqerr( ENOEXIT, "csp_xit: exit_seg" );
				return;
				}

			mem_wr_w( WORD_ADDR( exit_mp, MS_IPC), exit_ipc( exit_jtab ) );
			exit_mp = MSCW_RD( exit_mp, MS_DYN );
			}
		}
		break;
	case CSP_UREAD:
		{
		word unit;
		word buffer;
		word length;
		word block;
		word flags;

		flags = pop_w();
		block = pop_w();
		length = pop_w();
		buffer = pop_w();
		unit = pop_w();
		syscom->iorslt = unit_read( unit, buffer, length, block, flags );
		}
		break;
	case CSP_UWRITE:
		{
		word unit;
		word buffer;
		word length;
		word block;
		word flags;

		flags = pop_w();
		block = pop_w();
		length = pop_w();
		buffer = pop_w();
		unit = pop_w();
		syscom->iorslt = unit_write( unit, buffer, length, block, flags );
		}
		break;
	case CSP_TIM:
		{
		struct timeval tv;

		if ( gettimeofday( &tv, NULL ) < 0 )
			{
			perror("gettimeofday");
			mem_wr_w( pop_w(), 0 );
			mem_wr_w( pop_w(), 0 );
			syscom->lotime = 0;
			syscom->hitime = 0;
			}
		else
			{
			tv.tv_sec = ( tv.tv_usec * 60 / 1000000 ) + tv.tv_sec * 60;
			mem_wr_w( pop_w(),    ( tv.tv_sec >> 0 ) & 0xffff );
			syscom->lotime = ( tv.tv_sec >> 0 ) & 0xffff;

			mem_wr_w( pop_w(),    ( tv.tv_sec >> 16 ) & 0xffff );
			syscom->hitime = ( tv.tv_sec >> 16 ) & 0xffff;
			}
		}
		break;
	case CSP_IDS:
		{
		word reserved_word_ptr;
		word sym_buffer_ptr;

		reserved_word_ptr = pop_w();
		sym_buffer_ptr = pop_w();
		id_search( reserved_word_ptr, sym_buffer_ptr );
		}
		break;
	case CSP_TRS:
		{
		word token_tree_ptr;
		word result_ptr;
		word node_ptr;

		token_tree_ptr = pop_w();
		result_ptr = pop_w();
		node_ptr = pop_w();
		push_w( tree_search( token_tree_ptr, result_ptr, node_ptr ) );
		}
		break;
	case CSP_FLC:
		{
		word ch;
		word len;
		word addr;

		ch   = pop_w();
		len  = pop_w();
		addr = pop_w();
		if ( ! ( len & 0x8000 ) )
			while ( len )
				{
				mem_wr_b( addr, ch );
				len--;
				addr++;
				}
		}
		break;
	case CSP_SCN:
		{
#define SCAN_FOR	( ! past )
#define SCAN_PAST	( past )

		word mask;
		word start;
		word ch;
		word past;
		word limit;
		word res;

		mask  = pop_w();
		start = pop_w();
		ch    = pop_w();
		past  = pop_w();
		limit = pop_w();
		if ( limit & 0x8000 )
			{ /* down from start */
			limit = 0x10000 - limit;
			for ( res = 0; res < limit; res++ )
				if ( mem_rd_b( start - res ) == ch )
					{
					if ( SCAN_FOR )
						break;
					}
				else
					{
					if ( SCAN_PAST )
						break;
					}
			push_w( 0x10000 - res );
			}
		else
			{ /* up from start */
			for ( res = 0; res < limit; res++ )
				if ( mem_rd_b( start + res ) == ch )
					{
					if ( SCAN_FOR )
						break;
					}
				else
					{
					if ( SCAN_PAST )
						break;
					}
			push_w( res );
			}
		}
		break;
	case CSP_USTAT:
		{
		word unit;
		word buffer;
		word length;
		word flags;

		flags  = pop_w();
		length = pop_w();
		buffer = pop_w();
		unit   = pop_w();
		syscom->iorslt = unit_stat( unit, buffer, length, flags );
		}
		break;
	case CSP_LDSEG:
		load_segment( pop_w() );
		break;
	case CSP_ULDSEG:
		unload_segment( pop_w() );
		break;
	case CSP_TRC:
		{
		real value;

		value = pop_r();
		if ( value < 0 )
			push_w( ceil( value ) );
		else
			push_w( floor( value ) );
		}
		break;
	case CSP_RND:
		push_w( rint( pop_r() ) );
		break;
	case CSP_SIN:
		push_r( sin( pop_r() ) );
		break;
	case CSP_COS:
		push_r( cos( pop_r() ) );
		break;
	case CSP_LOG:
		push_r( log10( pop_r() ) );
		break;
	case CSP_ATAN:
		push_r( atan( pop_r() ) );
		break;
	case CSP_LN:
		push_r( log( pop_r() ) );
		break;
	case CSP_EXP:
		push_r( exp( pop_r() ) );
		break;
	case CSP_SQT:
		push_r( sqrt( pop_r() ) );
		break;
	case CSP_MRK:
		clear_gdirp();
		mem_wr_w( pop_w(), np );
		break;
	case CSP_RLS:
		np = mem_rd_w( pop_w() );
		stack_check();
		syscom->gdirp = NIL;
		break;
	case CSP_IOR:
		push_w( syscom->iorslt );
		break;
	case CSP_UBUSY:
		push_w( unit_busy( pop_w() ) );
		break;
	case CSP_POT:
		{
		int value;
		float pot_table[] =
			{
			1e0,  1e1,  1e2,  1e3,  1e4,
			1e5,  1e6,  1e7,  1e8,  1e9,
			1e10, 1e11, 1e12, 1e13, 1e14,
			1e15, 1e16, 1e17, 1e18, 1e19,
			1e20, 1e21, 1e22, 1e23, 1e24,
			1e25, 1e26, 1e27, 1e28, 1e29,
			1e30, 1e31, 1e32, 1e33, 1e34,
			1e35, 1e36, 1e37, 1e38, 1e39
			};

		value = pop_w();
		if ( ( value < 0 ) || ( value > 39 ) )
			xeqerr( EINVNDX, "csp_pot" );
		else
			push_r( pot_table[ value ] );
		}
		break;
	case CSP_UWAIT:
		syscom->iorslt = unit_wait( pop_w() );
		break;
	case CSP_UCLEAR:
		syscom->iorslt = unit_clear( pop_w() );
		break;
	case CSP_HLT:
		return;
		break;
	case CSP_MAV:
		{
		word free_bytes;

		if ( syscom->gdirp )
			free_bytes = sp - syscom->gdirp;
		else
			free_bytes = sp - np;
		push_w( free_bytes / 2 );
		}
		break;
	default:
		xeqerr( ENOTIMP, "csp" );
		break;
	}
}

/* ========================================================================== */
/* procedure/function return p-codes */

/* -------------------------------------------------------------------------- */
/* RNP        return to non-base procedure */
void rnp()
{
ret( fetch_b() );
}

/* -------------------------------------------------------------------------- */
/* RBP        return to base procedure */
void rbp()
{
base = mem_rd_w( mp - 2 );
syscom->stkbase = base;

if ( ( base < sp ) || ( base > basemp ) )
	{
	xeqerr( ENOEXIT, "rbp" );
	return;
	}

ret( fetch_b() );
}

/* ========================================================================== */
/* set p-codes */

/* -------------------------------------------------------------------------- */
/* DIF        set difference */
void dif()
{
word size;
int index;
set_t set;

pop_s( &set );
size = pop_w();
if ( size > set.size )
	adj_s( &set, size );

index = 0;
while ( index < size )
	{
	set.data[ index ] = pop_w() & ~set.data[ index ];
	index++;
	}
while ( index < set.size )
	{
	set.data[ index ] = 0;
	index++;
	}

push_s( &set );
}

/* -------------------------------------------------------------------------- */
/* INN        set membership */
void inn()
{
word size;
word addr;
word val;

size = pop_w();
addr = sp;
sp = WORD_ADDR( sp, size );
val = pop_w();
if ( val >= ( WORD_SZ * size ) )
	push_w( boolean( 0 ) );
else
	push_w( boolean( mem_rd_w( WORD_ADDR( addr, ( val / WORD_SZ ) ) ) & ( 1 << ( val % WORD_SZ ) ) ) );
}

/* -------------------------------------------------------------------------- */
/* INT        set intersection */
void intt()
{
word size;
int index;
set_t set;

pop_s( &set );
size = pop_w();
if ( size > set.size )
	adj_s( &set, size );

index = 0;
while ( index < size )
	{
	set.data[ index ] &= pop_w();
	index++;
	}
while( index < set.size )
	{
	set.data[ index ] = 0;
	index++;
	}

push_s( &set );
}

/* -------------------------------------------------------------------------- */
/* SRS        build subrange set */
void srs()
{
signed_word j;
signed_word i;
word addr;
int size;
int count;

j = pop_w();
i = pop_w();
if ( ( j < SET_SZ ) && ( i < SET_SZ) && ( i >= 0 ) )
	{
	if ( i > j )
		push_w( 0 );
	else
		{
		size = ( j + WORD_SZ ) / WORD_SZ;
		for ( count = 0; count < size; count++ )
			push_w( 0 );
		while ( i <= j )
			{
			addr = WORD_ADDR( sp, i / WORD_SZ );
			mem_wr_w( addr, mem_rd_w( addr ) | ( 1 << i % WORD_SZ));
			i++;
			}
		push_w( size );
		}
	}
else
	xeqerr( EINVNDX, "srs" );
}

/* -------------------------------------------------------------------------- */
/* SGS        build singleton set */
void sgs()
{
word tos;
word addr;
int size;
int count;

tos = pop_w();
if ( tos < SET_SZ )
	{
	size = ( tos + WORD_SZ ) / WORD_SZ;
	for ( count = 0; count < size; count++ )
		push_w( 0 );
	addr = WORD_ADDR( sp, tos / WORD_SZ );
	mem_wr_w( addr, mem_rd_w( addr ) | ( 1 << tos % WORD_SZ ) );
	push_w( size );
	}
else
	xeqerr( EINVNDX, "sgs" );
}

/* -------------------------------------------------------------------------- */
/* UNI        set union */
void uni()
{
word  size;
int   count;
set_t set;

pop_s( &set );
size = pop_w();
if ( size > set.size )
	adj_s( &set, size );

for ( count = 0; count < size; count++ )
	set.data[ count ] |= pop_w();
push_s( &set );
}

/* -------------------------------------------------------------------------- */
/* ADJ        adjust set */
void adj()
{
word final_size;
word orig_size;
set_t set;

final_size = fetch_b();
orig_size = mem_rd_w( sp );
if ( final_size != orig_size )
	{
	pop_s( &set );
	adj_s( &set, final_size );
	push_s( &set );
	}
if ( final_size != pop_w() )
	xeqerr( EADJFAIL, "adj" );
}

/* ========================================================================== */
/* integer comparison p-codes */

/* -------------------------------------------------------------------------- */
/* EQUI       integer '=' */
void equi()
{
word tos;

tos = pop_w();
push_w( boolean( (signed_word) pop_w() == (signed_word) tos ) );
}

/* -------------------------------------------------------------------------- */
/* GEQI       '>=' integer */
void geqi()
{
word tos;

tos = pop_w();
push_w( boolean( (signed_word) pop_w() >= (signed_word) tos ) );
}

/* -------------------------------------------------------------------------- */
/* GTRI       '>' integer */
void grti()
{
word tos;

tos = pop_w();
push_w( boolean( (signed_word) pop_w() > (signed_word) tos ) );
}

/* -------------------------------------------------------------------------- */
/* LEGI       '<=' integer */
void leqi()
{
word tos;

tos = pop_w();
push_w( boolean( (signed_word) pop_w() <= (signed_word) tos ) );
}

/* -------------------------------------------------------------------------- */
/* LESI       '<' integer */
void lesi()
{
word tos;

tos = pop_w();
push_w( boolean( (signed_word) pop_w() < (signed_word) tos ) );
}

/* -------------------------------------------------------------------------- */
/* NEQI       '<>' integer */
void neqi()
{
word tos;

tos = pop_w();
push_w( boolean( (signed_word) pop_w() != (signed_word) tos ) );
}

/* ========================================================================== */
/* general comparison p-codes */

/* -------------------------------------------------------------------------- */
/* EQU        '=' for reals,strings,sets,byte arrays */
void equ()
{
word word_val;
real real_val;
set_t set1;
set_t set2;

switch( fetch_b() )
	{
	case( CMP_REAL ) :
		real_val = pop_r();
		push_w( boolean( pop_r() == real_val ));
		break;
	case( CMP_STR ) :
		word_val = pop_w();
		push_w( boolean( cmp_str( pop_w(), word_val ) == 0 ) );
		break;
	case( CMP_BOOL ) :
		push_w( boolean( ( pop_w() & 1 ) == ( pop_w() & 1 ) ) );
		break;
	case( CMP_SET ) :
		pop_s( &set1 );
		pop_s( &set2 );
		push_w( boolean( cmp_set( &set1, &set2 ) == 0 ) );
		break;
	case( CMP_BYTES ) :
		word_val = pop_w();
		push_w( boolean( cmp_bytes( pop_w(), word_val, fetch_b() ) == 0 ) );
		break;
	case( CMP_WORDS ) :
		word_val = pop_w();
		push_w( boolean( cmp_words( pop_w(), word_val, fetch_b() ) == 0 ) );
		break;
	default:
		xeqerr( ENOTIMP, "equ" );
		break;
	}
}

/* -------------------------------------------------------------------------- */
/* GEQ        '>=' for reals,strings,sets,byte arrays */
void geq()
{
word word_val;
real real_val;
set_t set1;
set_t set2;

switch( fetch_b() )
	{
	case( CMP_REAL ) :
		real_val = pop_r();
		push_w( boolean( pop_r() >= real_val ) );
		break;
	case( CMP_STR ) :
		word_val = pop_w();
		push_w( boolean( cmp_str( pop_w(), word_val ) >= 0 ) );
		break;
	case( CMP_BOOL ) :
		word_val = pop_w() & 1;
		push_w( boolean( ( pop_w() & 1 ) >= word_val ) );
		break;
	case( CMP_SET ) :
		pop_s(&set1);
		pop_s(&set2);
		push_w( boolean( is_subset( &set2, &set1 ) ) );
		break;
	case( CMP_BYTES ) :
		word_val = pop_w();
		push_w( boolean( cmp_bytes( pop_w(), word_val, fetch_b() ) >= 0 ) );
		break;
	default:
		xeqerr( ENOTIMP, "geq" );
		break;
	}
}

/* -------------------------------------------------------------------------- */
/* GTR        '>' for reals,strings,sets,byte arrays */
void grt()
{
word word_val;
real real_val;

switch( fetch_b() )
	{
	case( CMP_REAL ) :
		real_val = pop_r();
		push_w( boolean( pop_r() > real_val ) );
		break;
	case( CMP_STR ) :
		word_val = pop_w();
		push_w( boolean( cmp_str( pop_w(), word_val ) > 0 ) );
		break;
	case( CMP_BOOL ) :
		word_val = pop_w() & 1;
		push_w( boolean( ( pop_w() & 1 ) > word_val ) );
		break;
	case( CMP_BYTES ) :
		word_val = pop_w();
		push_w( boolean( cmp_bytes( pop_w(), word_val, fetch_b() ) > 0 ) );
		break;
	default:
		xeqerr( ENOTIMP, "grt" );
		break;
	}
}

/* -------------------------------------------------------------------------- */
void leq()
/* LEQ        '<=' for reals,strings,sets,byte arrays */
{
word word_val;
real real_val;
set_t set1;
set_t set2;

switch( fetch_b() )
	{
	case( CMP_REAL ) :
		real_val = pop_r();
		push_w( boolean( pop_r() <= real_val ) );
		break;
	case( CMP_STR ) :
		word_val = pop_w();
		push_w( boolean( cmp_str( pop_w(), word_val ) <= 0 ) );
		break;
	case( CMP_BOOL ) :
		word_val = pop_w() & 1;
		push_w( boolean( ( pop_w() & 1 ) <= word_val ) );
		break;
	case( CMP_SET ) :
		pop_s(&set1);
		pop_s(&set2);
		push_w( boolean( is_subset( &set1, &set2 ) ) );
		break;
	case( CMP_BYTES ) :
		word_val = pop_w();
		push_w( boolean( cmp_bytes( pop_w(), word_val, fetch_b() ) <= 0 ) );
		break;
	default:
		xeqerr( ENOTIMP, "leq" );
		break;
	}
}

/* -------------------------------------------------------------------------- */
/* LES        '<' for reals,strings,sets,byte arrays */
void les()
{
word word_val;
real real_val;

switch( fetch_b() )
	{
	case( CMP_REAL ) :
		real_val = pop_r();
		push_w( boolean( pop_r() < real_val ) );
		break;
	case( CMP_STR ) :
		word_val = pop_w();
		push_w( boolean( cmp_str( pop_w(), word_val ) < 0 ) );
		break;
	case( CMP_BOOL ) :
		word_val = pop_w() & 1;
		push_w( boolean( ( pop_w() & 1 ) < word_val ) );
		break;
	case( CMP_BYTES ) :
		word_val = pop_w();
		push_w( boolean( cmp_bytes( pop_w(), word_val, fetch_b() ) < 0 ) );
		break;
	default:
		xeqerr( ENOTIMP, "les" );
		break;
	}
}

/* -------------------------------------------------------------------------- */
/* NEQ        '<>' for reals,strings,sets,byte arrays */
void neq()
{
word word_val;
real real_val;
set_t set1;
set_t set2;

switch( fetch_b() )
	{
	case( CMP_REAL ) :
		real_val = pop_r();
		push_w( boolean( pop_r() != real_val ) );
		break;
	case( CMP_STR ) :
		word_val = pop_w();
		push_w( boolean( cmp_str( pop_w(), word_val ) != 0 ) );
		break;
	case( CMP_BOOL ) :
		push_w( boolean( ( pop_w() & 1 ) != ( pop_w() & 1 ) ) );
		break;
	case( CMP_SET ) :
		pop_s( &set1 );
		pop_s( &set2 );
		push_w( boolean( cmp_set( &set1, &set2 ) != 0 ) );
		break;
	case( CMP_BYTES ) :
		word_val = pop_w();
		push_w( boolean( cmp_bytes( pop_w(), word_val, fetch_b() ) != 0 ) );
		break;
	case( CMP_WORDS ) :
		word_val = pop_w();
		push_w( boolean( cmp_words( pop_w(), word_val, fetch_b() ) != 0 ) );
		break;
	default:
		xeqerr( ENOTIMP, "neq" );
		break;
	}
}

/* ========================================================================== */
/* jumps p-codes */

/* -------------------------------------------------------------------------- */
/* FJP        false jump */
void fjp()
{
word disp;

disp = fetch_b();
if ( ! ( pop_w() & 1 ) )
	ipc = jump( (signed_byte) disp );
}

/* -------------------------------------------------------------------------- */
/* EFJ        equal false jump */
void efj()
{
xeqerr( ENOTIMP, "efj" );
/*
word disp;

disp = fetch_b();
if ( pop_w() != pop_w() )
	ipc = jump( (signed_byte) disp );
*/
}

/* -------------------------------------------------------------------------- */
/* NFJ        not equal false jump */
void nfj()
{
xeqerr( ENOTIMP, "nfj" );
/*
word disp;

disp = fetch_b();
if ( pop_w() == pop_w() )
	ipc = jump( (signed_byte) disp );
*/
}

/* -------------------------------------------------------------------------- */
/* XJP        case jump */
void xjp()
{
word min;
word max;
word val;

ipc = WORD_ALIGN( ipc );
min = fetch_w();
max = fetch_w();
val = pop_w();
if ( ( val >= min ) && ( val <= max ) )
	{
	ipc = ipc + ( 2 * ( val - min ) ) + 2;
	ipc -= mem_rd_w( ipc );
	}
}

/* -------------------------------------------------------------------------- */
/* UJP        unconditional jump */
void ujp()
{
ipc = jump( (signed_byte) fetch_b() );
}

/* ========================================================================== */
/* string p-codes */

/* -------------------------------------------------------------------------- */
/* S1P        string to packed conversion */
void s1p()
{
push_w( pop_w() + 1 );
}

/* -------------------------------------------------------------------------- */
/* LCA        load constant string address */
void lca()
{
word length;

push_w( ipc );
length = fetch_b();
ipc += length;
}

/* -------------------------------------------------------------------------- */
/* IXS        index string array */
void ixs()
{
word index;
word str_ptr;

index = pop_w();
str_ptr = pop_w();
if ( ( index < 1 ) || ( index > MAX_STR_LEN ) || ( index > mem_rd_b( str_ptr ) ) )
	{
	push_w( str_ptr + 1 );
	xeqerr( EINVNDX, "ixs" );
	return;
	}
push_w( str_ptr + index );
}

/* -------------------------------------------------------------------------- */
/* SAS        string assign */
void sas()
{
byte src_len;
byte dst_len;
word src_addr;
word dst_addr;

dst_len = fetch_b();
src_addr = pop_w();
dst_addr = pop_w();
if ( src_addr & 0xff00 )
	{	/* copy the string from src_addr to dst_addr */
	src_len = mem_rd_b( src_addr + 0 );
	if ( src_len > dst_len )
		{
		xeqerr( ES2LONG, "sas" );
		return;
		}
	mv_left( dst_addr, src_addr, src_len + 1 );
	}
else
	{	/* store the char TOS in dst_addr */
	mem_wr_b( dst_addr + 0, 1 );		/* make string of len 1		    */
	mem_wr_b( dst_addr + 1, src_addr );	/* containing char on stack	    */
	}
}

/* -------------------------------------------------------------------------- */
/* S2P        string to packed conversion */
void s2p()
{
word tos;

tos = pop_w();
push_w( pop_w() + 1 );
push_w( tos );
}

/* ========================================================================== */
/* indirect, records, arrays, and indexing */

/* -------------------------------------------------------------------------- */
/* IXB        index byte array */
void ixb()
{
push_w( pop_w() +  pop_w() );
}

/* -------------------------------------------------------------------------- */
/* IND        static index and load word */
void ind()
{
push_w( mem_rd_w( WORD_ADDR( pop_w(), get_big() ) ) );
}

/* -------------------------------------------------------------------------- */
/* SIND 0..7  short index and load word */
void sind( word opcode )
{
push_w( mem_rd_w( WORD_ADDR( pop_w(), opcode - SIND0 ) ) );
}

/* -------------------------------------------------------------------------- */
/* INC        increment field pointer */
void inc()
{
push_w( BYTE_ADDR( pop_w(), get_big() ) );
}

/* -------------------------------------------------------------------------- */
/* STO        store indirect */
void sto()
{
word tos;

tos = pop_w();
mem_wr_w( pop_w(), tos );
}

/* -------------------------------------------------------------------------- */
/* IXA        index array */
void ixa()
{
word index;

index = pop_w();
push_w( WORD_ADDR( pop_w(), index * get_big() ) );
}

/* -------------------------------------------------------------------------- */
/* MOV        move words */
void mov()
{
word len;
word src_addr;
word dst_addr;

len = get_big();
src_addr = pop_w();
dst_addr = pop_w();
while ( len )
	{
	mem_wr_w( dst_addr, mem_rd_w( src_addr ) );
	len--;
	dst_addr = WORD_ADDR( dst_addr, 1 );
	src_addr = WORD_ADDR( src_addr, 1 );
	}
}

/* ========================================================================== */
/*  char vars, and byte array vars */

/* -------------------------------------------------------------------------- */
/* LDB        load byte */
void ldb()
{
push_w( mem_rd_b( pop_w() ) );
}

/* -------------------------------------------------------------------------- */
/* STB        store byte */
void stb()
{
word value;
word dst_addr;

value = pop_w();
dst_addr = pop_w();
mem_wr_b( dst_addr, value );
}

/* -------------------------------------------------------------------------- */
/* MVB        move bytes */
void mvb()
{
word len;
word src_addr;
word dst_addr;

len = get_big();
src_addr = pop_w();
dst_addr = pop_w();
mv_left( dst_addr, src_addr, len );
}

/* ========================================================================== */
/* packed array & record p-codes */

/* -------------------------------------------------------------------------- */
/* IXP        index packed array */
void ixp()
{
word elements_per_word;
word bits_per_element;
word index;

elements_per_word = fetch_b();
bits_per_element = fetch_b();
index = pop_w();

push_w( WORD_ADDR( pop_w(), index / elements_per_word ) );
push_w( bits_per_element );
push_w( ( index % elements_per_word ) * bits_per_element );
}

/* -------------------------------------------------------------------------- */
/* LDP        load packed field */
void ldp()
{
word offset;
word size;
word addr;

offset = pop_w() & 0xff;
size = pop_w() & 0xff;
addr = pop_w();
if ( offset + size > WORD_SZ )
	{
	xeqerr( EINVNDX, "ldp" );
	return;
	}
push_w( ( mem_rd_w( addr ) >> offset ) & ( ( 1 << size ) - 1 ) );
}

/* -------------------------------------------------------------------------- */
/* STP        store packed field */
void stp()
{
word word_val;
word offset;
word size;
word addr;

word_val = pop_w();
offset = pop_w() & 0xff;
size = pop_w();
addr = pop_w();
if ( offset + size > WORD_SZ )
	{
	xeqerr( EINVNDX, "stp" );
	return;
	}
word_val &= ( 1 << size ) - 1 ;
mem_wr_w( addr, ( mem_rd_w( addr ) & ~( ( ( 1 << size ) - 1 ) << offset ) ) | ( word_val << offset ) );
}

/* ========================================================================== */
/* misc p-codes */

/* BYT        byte conversion */
/* NOP        no operation */

/* -------------------------------------------------------------------------- */
/* BPT        breakpoint */
void bpt()
{
word pc;

pc = get_big();
if (	( syscom->bugstate >= 3) ||
	( pc == syscom->brkpts[ 0 ] ) ||
	( pc == syscom->brkpts[ 1 ] ) ||
	( pc == syscom->brkpts[ 2 ] ) ||
	( pc == syscom->brkpts[ 3 ] )
	)
	xeqerr( EBRKPNT, "bpt" );
}

/* ========================================================================== */
/* xeqerror & iloop */

/* -------------------------------------------------------------------------- */
void xeqerr( word err, char *detail )
{
static int xeqerr_recursion = 0;
word new_seg;

fprintf( stderr, "exqerr: %d, %s\n", err, detail);

if ( xeqerr_recursion > 0 )
	{
	return;
	}
xeqerr_recursion++;

new_seg = int_seg_tbl[ 0 ].code_addr + syscom->segtbl[ 0 ].code_len - 2;
syscom->xeqerr = err;
syscom->bomipc = ipc;
sp += -14; /* size of execerror stack + temp */
call( new_seg, 2, basemp, 0 ); /* call execerror => seg 0, proc 2 */
syscom->bombp  = mp;

xeqerr_recursion--;
}

/* -------------------------------------------------------------------------- */
void interpret()
{
byte opcode;

seg     = syscom->seg;
mp      = syscom->lastmp;
basemp  = syscom->lastmp;
base    = syscom->stkbase;
jtab    = syscom->jtab;
ipc     = entry_ipc( jtab );

while ( 1 )
	{
	opcode = fetch_b();
	if ( opcode < 128 ) /* SLDC */
		push_w( opcode );
	else
		switch ( opcode )
			{
			case( 128 ) : /* ABI  */ abi();  break;
			case( 129 ) : /* ABR  */ abr();  break;
			case( 130 ) : /* ADI  */ adi();  break;
			case( 131 ) : /* ADR  */ adr();  break;
			case( 132 ) : /* AND  */ and();  break;
			case( 133 ) : /* DIF  */ dif();  break;
			case( 134 ) : /* DVI  */ dvi();  break;
			case( 135 ) : /* DVR  */ dvr();  break;
			case( 136 ) : /* CHK  */ chk();  break;
			case( 137 ) : /* FLO  */ flo();  break;
			case( 138 ) : /* FLT  */ flt();  break;
			case( 139 ) : /* INN  */ inn();  break;
			case( 140 ) : /* INT  */ intt(); break;
			case( 141 ) : /* IOR  */ ior();  break;
			case( 142 ) : /* MODI */ modi(); break;
			case( 143 ) : /* MPI  */ mpi();  break;
			case( 144 ) : /* MPR  */ mpr();  break;
			case( 145 ) : /* NGI  */ ngi();  break;
			case( 146 ) : /* NGR  */ ngr();  break;
			case( 147 ) : /* NOT  */ not();  break;
			case( 148 ) : /* SRS  */ srs();  break;
			case( 149 ) : /* SBI  */ sbi();  break;
			case( 150 ) : /* SBR  */ sbr();  break;
			case( 151 ) : /* SGS  */ sgs();  break;
			case( 152 ) : /* SQI  */ sqi();  break;
			case( 153 ) : /* SQR  */ sqr();  break;
			case( 154 ) : /* STO  */ sto();  break;
			case( 155 ) : /* IXS  */ ixs();  break;
			case( 156 ) : /* UNI  */ uni();  break;
			case( 157 ) : /* S2P  */ s2p();  break;
			case( 158 ) : /* CSP  */ csp();  break;
			case( 159 ) : /* LDCN */ ldcn(); break;
			case( 160 ) : /* ADJ  */ adj();  break;
			case( 161 ) : /* FJP  */ fjp();  break;
			case( 162 ) : /* INC  */ inc();  break;
			case( 163 ) : /* IND  */ ind();  break;
			case( 164 ) : /* IXA  */ ixa();  break;
			case( 165 ) : /* LAO  */ lao();  break;
			case( 166 ) : /* LCA  */ lca();  break;
			case( 167 ) : /* LDO  */ ldo();  break;
			case( 168 ) : /* MOV  */ mov();  break;
			case( 169 ) : /* MVB  */ mvb();  break;
			case( 170 ) : /* SAS  */ sas();  break;
			case( 171 ) : /* SRO  */ sro();  break;
			case( 172 ) : /* XJP  */ xjp();  break;
			case( 173 ) : /* RNP  */ rnp();  break;
			case( 174 ) : /* CIP  */ cip();  break;
			case( 175 ) : /* EQU  */ equ();  break;
			case( 176 ) : /* GEQ  */ geq();  break;
			case( 177 ) : /* GRT  */ grt();  break;
			case( 178 ) : /* LDA  */ lda();  break;
			case( 179 ) : /* LDC  */ ldc();  break;
			case( 180 ) : /* LEQ  */ leq();  break;
			case( 181 ) : /* LES  */ les();  break;
			case( 182 ) : /* LOD  */ lod();  break;
			case( 183 ) : /* NEQ  */ neq();  break;
			case( 184 ) : /* STR  */ str();  break;
			case( 185 ) : /* UJP  */ ujp();  break;
			case( 186 ) : /* LDP  */ ldp();  break;
			case( 187 ) : /* STP  */ stp();  break;
			case( 188 ) : /* LDM  */ ldm();  break;
			case( 189 ) : /* STM  */ stm();  break;
			case( 190 ) : /* LDB  */ ldb();  break;
			case( 191 ) : /* STB  */ stb();  break;
			case( 192 ) : /* IXP  */ ixp();  break;
			case( 193 ) : /* RBP  */ rbp();  break;
			case( 194 ) : /* CBP  */ cbp();  break;
			case( 195 ) : /* EQUI */ equi(); break;
			case( 196 ) : /* GEQI */ geqi(); break;
			case( 197 ) : /* GRTI */ grti(); break;
			case( 198 ) : /* LLA  */ lla();  break;
			case( 199 ) : /* LDCI */ ldci(); break;
			case( 200 ) : /* LEQI */ leqi(); break;
			case( 201 ) : /* LESI */ lesi(); break;
			case( 202 ) : /* LDL  */ ldl();  break;
			case( 203 ) : /* NEQI */ neqi(); break;
			case( 204 ) : /* STL  */ stl();  break;
			case( 205 ) : /* CXP  */ cxp();  break;
			case( 206 ) : /* CLP  */ clp();  break;
			case( 207 ) : /* CGP  */ cgp();  break;
			case( 208 ) : /* S1P  */ s1p();  break;
			case( 209 ) : /* IXB  */ ixb();  break;
			case( 210 ) : /* BYT  */         break;
			case( 211 ) : /* EFJ  */ efj();  break;
			case( 212 ) : /* NFJ  */ nfj();  break;
			case( 213 ) : /* BPT  */ bpt();  break;
			case( 214 ) : /* XIT  */ return; break;
			case( 215 ) : /* NOP  */         break;
			case( 216 ) : /* "SLDL0"  */
			case( 217 ) : /* "SLDL1"  */
			case( 218 ) : /* "SLDL2"  */
			case( 219 ) : /* "SLDL3"  */
			case( 220 ) : /* "SLDL4"  */
			case( 221 ) : /* "SLDL5"  */
			case( 222 ) : /* "SLDL6"  */
			case( 223 ) : /* "SLDL7"  */
			case( 224 ) : /* "SLDL8"  */
			case( 225 ) : /* "SLDL9"  */
			case( 226 ) : /* "SLDL10" */
			case( 227 ) : /* "SLDL11" */
			case( 228 ) : /* "SLDL12" */
			case( 229 ) : /* "SLDL13" */
			case( 230 ) : /* "SLDL14" */
			case( 231 ) : /* "SLDL15" */ sldl( opcode ); break;
			case( 232 ) : /* "SLDO0"  */
			case( 233 ) : /* "SLDO1"  */
			case( 234 ) : /* "SLDO2"  */
			case( 235 ) : /* "SLDO3"  */
			case( 236 ) : /* "SLDO4"  */
			case( 237 ) : /* "SLDO5"  */
			case( 238 ) : /* "SLDO6"  */
			case( 239 ) : /* "SLDO7"  */
			case( 240 ) : /* "SLDO8"  */
			case( 241 ) : /* "SLDO9"  */
			case( 242 ) : /* "SLDO10" */
			case( 243 ) : /* "SLDO11" */
			case( 244 ) : /* "SLDO12" */
			case( 245 ) : /* "SLDO13" */
			case( 246 ) : /* "SLDO14" */
			case( 247 ) : /* "SLDO15" */ sldo( opcode );  break;
			case( 248 ) : /* "SIND0"  */
			case( 249 ) : /* "SIND1"  */
			case( 250 ) : /* "SIND2"  */
			case( 251 ) : /* "SIND3"  */
			case( 252 ) : /* "SIND4"  */
			case( 253 ) : /* "SIND5"  */
			case( 254 ) : /* "SIND6"  */
			case( 255 ) : /* "SIND7"  */ sind( opcode ); break;
			default : xeqerr( ENOTIMP, "iloop" ); break;
			}
	}
}
